﻿using System;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.EngineCore;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    /// <summary>
    /// Class MapObjectOverlay displays a map object as overlay on the map view.
    /// </summary>
    class MapObjectOverlay
    {
        // The default outline color
        static private Color LINE_COLOR = Color.fromRGB(0, 0, 0);
        // The default fill color
        static private Color FILL_COLOR = Color.fromRGB(200, 200, 0);
        // The default alpha background color
        static private Color BACKGROUND_COLOR = Color.fromRGB(255, 255, 255);

        /** The layer */
        private MapObjectLayer layer;
        /** The map object */
        private MapObject mapObject;

        /** The fill graphic */
        private int fillGraphic;
        /** The outline graphic */
        private int outlineGraphic;
        /** The point graphic */
        private int pointGraphic;
        /** The label graphic */
        private int labelGraphic;

        /// <summary>
        /// Create a new map object overlay
        /// </summary>
        /// <param name="mapObject">the map object to display</param>
        internal MapObjectOverlay(MapObjectLayer layer,MapObject mapObject)
        {
            /* Initialize */
            this.layer = layer;
            this.mapObject = mapObject;
            this.fillGraphic = -1;
            this.outlineGraphic = -1;
            this.pointGraphic = -1;
            this.labelGraphic = -1;
        }

        /// <summary>
        /// Close the map object overlay
        /// </summary>
        internal void Close()
        {
            this.Clear();
        }

        /// <summary>
        /// Clear the map object overlay
        /// </summary>
        internal void Clear()
        {
            if (this.fillGraphic >= 0)
            {
                this.layer.getGraphics().Remove(this.fillGraphic);
                this.fillGraphic = -1;
            }
            if (this.outlineGraphic >= 0)
            {
                this.layer.getGraphics().Remove(this.outlineGraphic);
                this.outlineGraphic = -1;
            }
            if (this.pointGraphic >= 0)
            {
                this.layer.getGraphics().Remove(this.pointGraphic);
                this.pointGraphic = -1;
            }
            if (this.labelGraphic >= 0)
            {
                this.layer.getGraphics().Remove(this.labelGraphic);
                this.labelGraphic = -1;
            }
        }

        /// <summary>
        /// Change the visibility of the overlay
        /// </summary>
        /// <param name="visible"></param>
        internal void setVisible(bool visible)
        {
            if (this.fillGraphic >= 0) this.layer.getGraphics().SetVisible(this.fillGraphic, visible);
            if (this.outlineGraphic >= 0) this.layer.getGraphics().SetVisible(this.outlineGraphic, visible);
            if (this.pointGraphic >= 0) this.layer.getGraphics().SetVisible(this.pointGraphic, visible);
            if (this.labelGraphic >= 0) this.layer.getGraphics().SetVisible(this.labelGraphic, visible);
        }

        /// <summary>
        /// Scale a number of pixels
        /// </summary>
        /// <param name="pixelSize">the number of the pixels as defined by the 3D mapping viewer</param>
        /// <returns></returns>
        private static double scalePixelSize(double pixelSize)
        {
            // the 3D mapping viewer uses 96 dpi, but arcgis expects 72 dpi
            return pixelSize / 96.0 * 72.0;
        }

        /// <summary>
        /// Create a color object
        /// </summary>
        /// <param name="color">the color to create</param>
        /// <returns></returns>
        static private IRgbColor createColor(Color color,Color defaultColor)
        {
            if (color == null) color = defaultColor;
            IRgbColor color2 = new RgbColorClass();
            int alpha = (int)color.alpha; // alpha in colors does not work, so we assume a default background color and blend
            color2.Red = (int)(((255-alpha)*BACKGROUND_COLOR.red+(alpha)*color.red)/255);
            color2.Green = (int)(((255-alpha)*BACKGROUND_COLOR.green+(alpha)*color.green)/255);
            color2.Blue = (int)(((255-alpha)*BACKGROUND_COLOR.blue+(alpha)*color.blue)/255);
            //color2.Transparency = (byte)color.alpha; // this only provides on/off visibility 0 or 255, nothing in between
            return color2;
        }

        /// <summary>
        /// Create an overlay for an area object
        /// </summary>
        private void DrawArea(IDisplay display, bool movedView)
        {
            /* Already created? */
            if (this.fillGraphic >= 0) return;
            if (this.outlineGraphic >= 0) return;
            /* Create the geometry */
            IGeometry fillGeometry = MapObjects.getAreaGeometry(mapObject,false/*hasZ*/);
            if (fillGeometry == null) return;
            fillGeometry.Project(this.layer.getMap().SpatialReference);
            MapObject mapObject2 = new MapObject();
            mapObject2.spatialLine = this.mapObject.spatialArea;
            IGeometry outlineGeometry = MapObjects.getLineGeometry(mapObject2, false/*hasZ*/, true/*close*/);
            if (outlineGeometry == null) return;
            outlineGeometry.Project(this.layer.getMap().SpatialReference);
            /* Create the symbol */
            if (mapObject.fillColor == null) mapObject.fillColor = Color.fromRGBA(mapObject.color.red, mapObject.color.green, mapObject.color.blue, 128);
            IRgbColor lineColor = createColor(mapObject.color,LINE_COLOR);
            IRgbColor fillColor = createColor(mapObject.fillColor,FILL_COLOR);
            ISimpleLineSymbol outlineSymbol = new SimpleLineSymbolClass();
            outlineSymbol.Color = lineColor;
            outlineSymbol.Style = esriSimpleLineStyle.esriSLSSolid;
            outlineSymbol.Width = scalePixelSize(mapObject.lineWidth);
            ISimpleLineSymbol invisibleOutlineSymbol = new SimpleLineSymbolClass();
            invisibleOutlineSymbol.Style = esriSimpleLineStyle.esriSLSNull;
            invisibleOutlineSymbol.Width = 0.0;
            ISimpleFillSymbol fillSymbol = new SimpleFillSymbolClass();
            fillSymbol.Color = fillColor;
            fillSymbol.Style = esriSimpleFillStyle.esriSFSSolid;
            fillSymbol.Outline = invisibleOutlineSymbol;
            //Message.WriteLine("created area (" + this.geometry.SpatialReference.FactoryCode + ") for " + this.mapObject.objectId);
            // add the graphics
            IGraphicTracker graphics = this.layer.getGraphics();
            if (mapObject.fillColor.alpha > 100)
            {
                this.fillGraphic = graphics.Add(fillGeometry, graphics.CreateSymbol(fillSymbol as ISymbol, null/*3D*/));
                //graphics.SetTransparency(this.fillGraphic, 255 - mapObject.fillColor.alpha); // individual transparency does not work, many users have reported this
            }
            this.outlineGraphic = graphics.Add(outlineGeometry, graphics.CreateSymbol(outlineSymbol as ISymbol, null/*3D*/));
        }

        /// <summary>
        /// Create an overlay for a line object
        /// </summary>
        private void DrawLine(IDisplay display, bool movedView)
        {
            /* Already created? */
            if (this.outlineGraphic >= 0) return;
            /* Create the geometry */
            IGeometry outlineGeometry = MapObjects.getLineGeometry(mapObject, false/*hasZ*/, false/*close*/);
            if (outlineGeometry == null) return;
            outlineGeometry.Project(this.layer.getMap().SpatialReference);
            /* Create the symbol */
            IRgbColor lineColor = createColor(mapObject.color, LINE_COLOR);
            ISimpleLineSymbol outlineSymbol = new SimpleLineSymbolClass();
            outlineSymbol.Color = lineColor;
            outlineSymbol.Style = esriSimpleLineStyle.esriSLSSolid;
            outlineSymbol.Width = scalePixelSize(mapObject.lineWidth);
            //Message.WriteLine("created line (" + this.geometry.SpatialReference.FactoryCode + ") for " + this.mapObject.objectId);
            // add the graphics
            IGraphicTracker graphics = this.layer.getGraphics();
            this.outlineGraphic = graphics.Add(outlineGeometry, graphics.CreateSymbol(outlineSymbol as ISymbol, null/*3D*/));
        }

        /// <summary>
        /// Create an overlay for a point object
        /// </summary>
        private void DrawPoint(IDisplay display, bool movedView)
        {
            /* Already created? */
            if (this.pointGraphic >= 0) return;
            /* Create the geometry */
            IGeometry pointGeometry = MapObjects.GetMapObjectGeometry(mapObject, display);
            if (pointGeometry == null) return;
            pointGeometry.Project(this.layer.getMap().SpatialReference);
            /* Create the symbol */
            double pointSize = 12.0;
            IRgbColor pointColor = createColor(mapObject.color, LINE_COLOR);
            ISimpleMarkerSymbol pointSymbol = new SimpleMarkerSymbolClass();
            pointSymbol.Color = pointColor;
            pointSymbol.Size = scalePixelSize(pointSize);
            pointSymbol.Style = esriSimpleMarkerStyle.esriSMSCircle;
            //Message.WriteLine("created point ("+p.X+","+p.Y+","+p.SpatialReference.FactoryCode+") for " + this.mapObject.objectId);
            // add the graphics
            IGraphicTracker graphics = this.layer.getGraphics();
            this.pointGraphic = graphics.Add(pointGeometry, graphics.CreateSymbol(pointSymbol as ISymbol, null/*3D*/));
        }

        /// <summary>
        /// Create an overlay for a point symbol
        /// </summary>
        private void DrawSymbol(IDisplay display, bool movedView)
        {
            /* Already created? */
            if ((this.fillGraphic >= 0) && (movedView == false)) return;
            if ((this.outlineGraphic >= 0) && (movedView == false)) return;
            /* Create the geometry */
            IGeometry symbolGeometry = MapObjects.GetSymbolPathGeometry(mapObject, display);
            if (symbolGeometry == null) return;
            /* Create the symbol */
            if (mapObject.fillColor == null) mapObject.fillColor = Color.fromRGBA(mapObject.color.red, mapObject.color.green, mapObject.color.blue, 128);
            // same colors and using transparency?
            if ((mapObject.color.red == mapObject.fillColor.red) &&
                (mapObject.color.green == mapObject.fillColor.green) &&
                (mapObject.color.blue == mapObject.fillColor.blue) &&
                (mapObject.color.alpha == mapObject.fillColor.alpha) &&
                (mapObject.color.alpha<255))
            {
                // make the outline opaque to have a difference with the fill color
                // platforms that support transparency do not need to do this
                mapObject.color.alpha = 255;
            }
            IRgbColor lineColor = createColor(mapObject.color, LINE_COLOR);
            IRgbColor fillColor = createColor(mapObject.fillColor, FILL_COLOR);
            ISimpleLineSymbol outlineSymbol = new SimpleLineSymbolClass();
            outlineSymbol.Color = lineColor;
            outlineSymbol.Style = esriSimpleLineStyle.esriSLSSolid;
            outlineSymbol.Width = scalePixelSize(mapObject.lineWidth);
            ISimpleLineSymbol invisibleOutlineSymbol = new SimpleLineSymbolClass();
            invisibleOutlineSymbol.Style = esriSimpleLineStyle.esriSLSNull;
            invisibleOutlineSymbol.Width = 0.0;
            ISimpleFillSymbol fillSymbol = new SimpleFillSymbolClass();
            fillSymbol.Color = fillColor;
            fillSymbol.Style = esriSimpleFillStyle.esriSFSSolid;
            fillSymbol.Outline = invisibleOutlineSymbol;
            //Message.WriteLine("created symbol (" + this.geometry.SpatialReference.FactoryCode + ") for " + this.mapObject.objectId);
            // add the graphics
            IGraphicTracker graphics = this.layer.getGraphics();
            if (mapObject.fillColor.alpha > 100)
            {
                if (this.fillGraphic < 0) this.fillGraphic = graphics.Add(symbolGeometry, graphics.CreateSymbol(fillSymbol as ISymbol, null/*3D*/));
                else graphics.Replace(this.fillGraphic, symbolGeometry, graphics.CreateSymbol(fillSymbol as ISymbol, null/*3D*/));
                //graphics.SetTransparency(this.fillGraphic, 255 - mapObject.fillColor.alpha); // individual transparency does not work, many users have reported this
            }
            if (this.outlineGraphic < 0) this.outlineGraphic = graphics.Add(symbolGeometry, graphics.CreateSymbol(outlineSymbol as ISymbol, null/*3D*/));
            else graphics.Replace(this.outlineGraphic, symbolGeometry, graphics.CreateSymbol(outlineSymbol as ISymbol, null/*3D*/));
        }

        /// <summary>
        /// get a font to display a label with
        /// </summary>
        /// <param name="fontName">the name of the font</param>
        /// <param name="fontSize">the size of the font</param>
        /// <returns></returns>
        static private stdole.IFontDisp getFont(string fontName,double fontSize)
        {
            stdole.IFontDisp font0 = (stdole.IFontDisp)new stdole.StdFont();
            font0.Name = fontName;
            font0.Size = (decimal)fontSize;
            return font0;
        }

        /// <summary>
        /// Create an overlay for a text label
        /// </summary>
        private void DrawLabel(IDisplay display, bool movedView)
        {
            /* Already created? */
            if (this.labelGraphic >= 0) return;
            /* Nothing to render? */
            if (this.mapObject.label == null) return;
            if (this.mapObject.label.Length == 0) return;
            /* Get the label position */
            IGeometry labelPosition = MapObjects.createPoint(MapObjects.GetMapObjectCentroid(this.mapObject),MapObjects.VIEWER_CRS);
            if (labelPosition == null) return;
            labelPosition.Project(this.layer.getMap().SpatialReference);
            /* Get the label properties */
            Coordinate labelOffset = mapObject.labelPosition;
            if (labelOffset == null) labelOffset = Coordinate.fromXY("pixel", 0.0, 0.0);
            Color labelColor = mapObject.labelColor;
            if (labelColor == null) labelColor = new Color(255, 255, 255, 255);
            double labelFontSize = mapObject.labelFontSize;
            if (labelFontSize < 1.0) labelFontSize = 12;
            /* Create the symbol */
            ISimpleMarkerSymbol pointSymbol = new SimpleMarkerSymbolClass();
            //pointSymbol.Color = pointColor;
            pointSymbol.Size = 0.1;
            pointSymbol.Style = esriSimpleMarkerStyle.esriSMSCircle;
            IRgbColor mapLabelColor = createColor(labelColor, LINE_COLOR);
            TextSymbolClass labelSymbol = new TextSymbolClass();
            labelSymbol.Color = mapLabelColor;
            labelSymbol.Text = mapObject.label;
            labelSymbol.Size = labelFontSize;
            labelSymbol.Font = getFont("Arial",labelFontSize);
            labelSymbol.HorizontalAlignment = esriTextHorizontalAlignment.esriTHACenter;
            labelSymbol.VerticalAlignment = esriTextVerticalAlignment.esriTVACenter;
            labelSymbol.XOffset = scalePixelSize(labelOffset.x);
            labelSymbol.YOffset = -scalePixelSize(labelOffset.y)-2.0; // text offset goes bottom-up
            //Message.WriteLine("created label (" + this.geometry.SpatialReference.FactoryCode + ") for " + this.mapObject.objectId);
            // add the graphics
            IGraphicTracker graphics = this.layer.getGraphics();
            this.labelGraphic = graphics.Add(labelPosition, graphics.CreateSymbol(pointSymbol as ISymbol, null/*3D*/));
            graphics.SetLabel(this.labelGraphic, mapObject.label);
            graphics.SetTextSymbol(this.labelGraphic, labelSymbol as ISimpleTextSymbol);
        }

        /// <summary>
        /// Draw the map overlay
        /// </summary>
        /// <param name="display">the display to overlay on</param>
        internal void draw(IDisplay display,bool movedView)
        {
            /* Area object? */
            if (mapObject.spatialArea != null) DrawArea(display, movedView);
            /* Line object? */
            else if (mapObject.spatialLine != null) DrawLine(display, movedView);
            /* Point object? */
            else if (mapObject.spatialPoint != null && mapObject.symbolPath == null) DrawPoint(display, movedView);
            /* Symbol object? */
            else if (mapObject.spatialPoint != null && mapObject.symbolPath != null) DrawSymbol(display, movedView);
            /* Label? */
            if (mapObject.spatialPoint != null && mapObject.label != null) DrawLabel(display, movedView);
        }
    }
}
