﻿using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.Geometry;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcmap
{
    class MapObjects
    {
        // Define the CRS that will be used to communicate with the 3D Mapping SDK (WGS84 lon-lat geographic EPSG:4326)
        static internal string CRS = orbitgt.mapping3d.sdk.viewer.Constants.CRS_WGS84;

        // the EPSG code of the coordinate reference system (CRS) we use to communicate with the viewer
        static internal int VIEWER_CRS_CODE = 4326; // the well-known WGS84 system
        static internal ISpatialReference VIEWER_CRS = null;

        /// <summary>
        /// Create a 2D point
        /// </summary>
        static internal IPoint createPoint2D(double x, double y, ISpatialReference crs)
        {
            IPoint point = new PointClass();
            point.SpatialReference = crs;
            point.X = x;
            point.Y = y;
            (point as IZAware).ZAware = false;
            return point;
        }

        /// <summary>
        /// Create a 3D point
        /// </summary>
        static internal IPoint createPoint3D(double x, double y, double z, ISpatialReference crs)
        {
            IPoint point = new PointClass();
            point.SpatialReference = crs;
            point.X = x;
            point.Y = y;
            point.Z = z;
            (point as IZAware).ZAware = true;
            return point;
        }

        /// <summary>
        /// Create a point from a coordinate
        /// </summary>
        static internal IPoint createPoint(Coordinate c, ISpatialReference crs)
        {
            if (c == null) return null;
            if (c.is3D) return createPoint3D(c.x, c.y, c.z, crs);
            else return createPoint2D(c.x, c.y, crs);
        }

        /// <summary>
        /// Create a point from a coordinate
        /// </summary>
        static internal IPoint createPointZ(Coordinate c, bool hasZ, ISpatialReference crs)
        {
            if (c == null) return null;
            double z = c.is3D ? c.z : 0.0;
            if (hasZ) return createPoint3D(c.x, c.y, z, crs);
            else return createPoint2D(c.x, c.y, crs);
        }

        /// <summary>
        /// Convert a point map object to a point geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal IGeometry getPointGeometry(MapObject mapObject, bool hasZ)
        {
            if (mapObject.spatialPoint == null) return null;
            Coordinate point0 = mapObject.spatialPoint;
            return createPointZ(point0, hasZ, VIEWER_CRS);
        }

        /// <summary>
        /// Convert a line map object to a line geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal IGeometry getLineGeometry(MapObject mapObject, bool hasZ, bool close)
        {
            if (mapObject.spatialLine == null) return null;
            IGeometryCollection polyline = new PolylineClass();
            ((IGeometry)polyline).SpatialReference = VIEWER_CRS;
            ((IZAware)polyline).ZAware = hasZ;
            for (int i=0; i<mapObject.spatialLine.parts.size(); i++)
            {
                CoordinatePart part0 = mapObject.spatialLine.parts.get(i);
                IPointCollection part = new PathClass();
                ((IGeometry)part).SpatialReference = VIEWER_CRS;
                ((IZAware)part).ZAware = hasZ;
                for (int j=0; j<part0.points.size(); j++) part.AddPoint(createPointZ(part0.points.get(j), hasZ, VIEWER_CRS));
                if (close)
                {
                    Coordinate first = part0.points.get(0);
                    Coordinate last = part0.points.get(part0.points.size() - 1);
                    bool closed = (first.x == last.x) && (first.y == last.y);
                    if (closed == false) part.AddPoint(createPointZ(part0.points.get(0), hasZ, VIEWER_CRS));
                }
                polyline.AddGeometry(part as IGeometry);
            }
            return polyline as IGeometry;
        }

        /// <summary>
        /// Convert an area map object to an area geometry
        /// </summary>
        /// <param name="mapObject"></param>
        /// <returns></returns>
        static internal IGeometry getAreaGeometry(MapObject mapObject,bool hasZ)
        {
            if (mapObject.spatialArea == null) return null;
            IGeometryCollection polygon = new PolygonClass();
            ((IGeometry)polygon).SpatialReference = VIEWER_CRS;
            ((IZAware)polygon).ZAware = hasZ;
            for (int i = 0; i < mapObject.spatialArea.parts.size(); i++)
            {
                CoordinatePart part0 = mapObject.spatialArea.parts.get(i);
                IPointCollection part = new RingClass();
                ((IGeometry)part).SpatialReference = VIEWER_CRS;
                ((IZAware)part).ZAware = hasZ;
                for (int j = 0; j < part0.points.size(); j++) part.AddPoint(createPointZ(part0.points.get(j), hasZ, VIEWER_CRS));
                polygon.AddGeometry(part as IGeometry);
            }
            return polygon as IGeometry;
        }

        /// <summary>
        /// Get the centroid position of a path.
        /// </summary>
        /// <param name="path">the path</param>
        /// <returns>the centroid position</returns>
        static public Coordinate GetPathCentroid(CoordinatePath path)
        {
            double tx = 0.0;
            double ty = 0.0;
            double tz = 0.0;
            int tcount = 0;
            foreach (CoordinatePart part in path.parts.toArray())
            {
                foreach (Coordinate point in part.points.toArray())
                {
                    tx += point.x;
                    ty += point.y;
                    if (point.is3D) tz += point.z;
                    tcount++;
                }
            }
            if (tcount == 0) return null;
            return new Coordinate(path.crs, tx / tcount, ty / tcount, tz / tcount, (tz != 0.0));
        }

        /// <summary>
        /// Get the centroid position of a map object.
        /// </summary>
        /// <param name="mapObject">the map object</param>
        /// <returns>the centroid position (null if not available)</returns>
        static public Coordinate GetMapObjectCentroid(MapObject mapObject)
        {
            if (mapObject.spatialPoint != null) return mapObject.spatialPoint;
            if (mapObject.spatialLine != null) return GetPathCentroid(mapObject.spatialLine);
            if (mapObject.spatialArea != null) return GetPathCentroid(mapObject.spatialArea);
            return null;
        }

        /// <summary>
        /// Get the position of a symbol path on the map view.
        /// </summary>
        /// <param name="mapObject">the symbol path object</param>
        /// <param name="display">the display to overlay on</param>
        /// <returns>the map point (in the map spatial reference)</returns>
        static public IPoint GetSymbolPathOrigin(MapObject mapObject, IDisplay display)
        {
            Coordinate centroid = GetMapObjectCentroid(mapObject);
            IPoint origin = new PointClass();
            origin.SpatialReference = VIEWER_CRS;
            origin.X = centroid.x;
            origin.Y = centroid.y;
            origin.Project(display.DisplayTransformation.SpatialReference);
            return origin;
        }

        /// <summary>
        /// Get the geometry of a symbol path on the map view
        /// </summary>
        /// <param name="mapObject">the symbol path object</param>
        /// <param name="display">the display to overlay on</param>
        /// <returns>the map geometry (in the map spatial reference)</returns>
        static public IGeometry GetSymbolPathGeometry(MapObject mapObject, IDisplay display)
        {
            /* Get the display position of the symbol origin */
            int originX = 0;
            int originY = 0;
            display.DisplayTransformation.FromMapPoint(GetSymbolPathOrigin(mapObject, display),out originX,out originY);
            /* Add the parts of the symbol path */
            IGeometryCollection symbol = new PolygonClass();
            ((IGeometry)symbol).SpatialReference = display.DisplayTransformation.SpatialReference;
            for (int i=0; i<mapObject.symbolPath.parts.size(); i++)
            {
                /* Add all points of the next part */
                CoordinatePart part0 = mapObject.symbolPath.parts.get(i);
                IPointCollection part = new RingClass();
                ((IGeometry)part).SpatialReference = display.DisplayTransformation.SpatialReference;
                for (int j=0; j<part0.points.size(); j++)
                {
                    /* Add the next point */
                    Coordinate pixel = part0.points.get(j);
                    IPoint pointM = display.DisplayTransformation.ToMapPoint(originX + (int)pixel.x, originY + (int)pixel.y);
                    part.AddPoint(pointM);
                }
                // line segment?
                if (part0.points.size() == 2)
                {
                    /* Add the next point */
                    Coordinate pixel = part0.points.get(0);
                    IPoint pointM = display.DisplayTransformation.ToMapPoint(originX + (int)pixel.x+1, originY + (int)pixel.y);
                    part.AddPoint(pointM);
                }
                //part.AddPoint(part.get_Point(0));
                symbol.AddGeometry(part as IGeometry);
            }
            /* Return the geometry */
            return symbol as IGeometry;
        }

        /// <summary>
        /// Get the geometry of a map object (must be called on the UI thread).
        /// </summary>
        /// <param name="mapObject">the map object</param>
        /// <param name="display">the display to overlay on</param>
        /// <returns>the map geometry</returns>
        static public IGeometry GetMapObjectGeometry(MapObject mapObject, IDisplay display)
        {
            /* Point? */
            if (mapObject.spatialPoint != null)
            {
                /* Symbol path? */
                if (mapObject.symbolPath != null)
                {
                    return GetSymbolPathGeometry(mapObject, display);
                }
                return getPointGeometry(mapObject, false/*hasZ*/);
            }
            /* Line? */
            if (mapObject.spatialLine != null)
            {
                return getLineGeometry(mapObject, false/*hasZ*/, false/*close*/);
            }
            /* Area? */
            if (mapObject.spatialArea != null)
            {
                return getAreaGeometry(mapObject,false/*hasZ*/);
            }
            /* Unknown type */
            return null;
        }

        /// <summary>
        /// Convert a map extent to another CRS (must be called on the UI thread)
        /// </summary>
        /// <param name="extent">the map extent</param>
        /// <param name="crs">the target CRS</param>
        /// <returns></returns>
        static public Bounds transformExtent(IEnvelope mapExtent, ISpatialReference crs)
        {
            ISpatialReference mapCRS = mapExtent.SpatialReference;
            if (mapCRS == null) return null;
            IPoint p0 = MapObjects.createPoint2D(mapExtent.XMin, mapExtent.YMin, mapCRS);
            p0.Project(crs);
            if (p0.IsEmpty) return null;
            IPoint p1 = MapObjects.createPoint2D(mapExtent.XMax, mapExtent.YMin, mapCRS);
            p1.Project(crs);
            if (p1.IsEmpty) return null;
            IPoint p2 = MapObjects.createPoint2D(mapExtent.XMin, mapExtent.YMax, mapCRS);
            p2.Project(crs);
            if (p2.IsEmpty) return null;
            IPoint p3 = MapObjects.createPoint2D(mapExtent.XMax, mapExtent.YMax, mapCRS);
            p3.Project(crs);
            if (p3.IsEmpty) return null;
            Bounds worldExtent = new Bounds("" + crs.FactoryCode);
            worldExtent.addXY(p0.X, p0.Y);
            worldExtent.addXY(p1.X, p1.Y);
            worldExtent.addXY(p2.X, p2.Y);
            worldExtent.addXY(p3.X, p3.Y);
            return worldExtent;
        }
    }
}
